"""
This file contains the scene description

This is an early prototype, more to come soon...
"""


class Scene(object):
    def __init__(self):
        self.camera = None
        self.nodes = []
        self.lights = []

    def add(self, object):
        self.nodes.append(object)

    def addLight(self, light):
        self.lights.append(light)

    def setCamera(self, cam):
        self.camera = cam



"""
This is the geometric object triangle

(renderable object)
"""

from ..geometry import Shape, Vertex
from ..material import PhongMaterial
from ..math import *


class Triangle(Shape):

    """Triangle class for raytracing"""

    def __init__(self, a: Vertex, b: Vertex, c: Vertex, material=PhongMaterial()):
        if type(a) != Vertex or type(b) != Vertex or type(c) != Vertex:
            raise ValueError("Please initialize Triangle with 3x Vertex")
        Shape.__init__(self, "Triangle")
        self.a = a
        self.b = b
        self.c = c
        self.material = material
        self.EPSILON = 0.000001

    def __str__(self):
        return "△ABC: Position[" + str(self.a.position) + ", " + str(self.b.position) + ", " + str(
            self.c.position) + "]"

    # ------------------------------------------------------------------------------------------------------------------
    def area(self) -> float:
        """
        Calculate area of triangle
        :return: area
        """
        return cross3(self.b.position - self.a.position,
                      self.c.position - self.a.position).length() / 2.0

    # ------------------------------------------------------------------------------------------------------------------
    def toBarycentric(self, p: Vec3) -> Vec3:
        """
        Calculate barycentric coordinate of point p
        :param P: the input point (cartesian)
        :return: barycentric coordinate
        """
        abc = triangleArea(self.a.position, self.b.position, self.c.position)
        pbc = triangleArea(p, self.b.position, self.c.position)
        apc = triangleArea(self.a.position, p, self.c.position)

        if abc == 0.0:
            return Vec3(0, 0, 0)

        x = pbc / abc
        y = apc / abc
        return Vec3(x, y, 1.0 - x - y)

    # ------------------------------------------------------------------------------------------------------------------
    def fromBarycentric(self, b: Vec3) -> Vec3:
        """
        Calculate cartesian coordinate from barycentric coordinate
        :param b: barycentric coordinate
        :return: cartesian coordinate
        """
        return self.a.position * b.x + self.b.position * b.y + self.c.position * b.z

    # ------------------------------------------------------------------------------------------------------------------
    def circumradius(self) -> float:
        a = (self.c.position - self.b.position).length()
        b = (self.c.position - self.a.position).length()
        c = (self.b.position - self.a.position).length()

        s = (a + b + c) / 2.0  # semiperimeter
        z = a * b * c
        n = 4.0 * math.sqrt(s * (a + b - s) * (a + c - s) * (b + c - s))
        if n == 0.0:
            return 0.0
        return z / n

    # ------------------------------------------------------------------------------------------------------------------
    def inradius(self) -> float:
        a = (self.c.position - self.b.position).length()
        b = (self.c.position - self.a.position).length()
        c = (self.b.position - self.a.position).length()
        s = (a + b + c) / 2.0
        if s == 0.0:
            return 0.0
        return math.sqrt((s - a) * (s - b) * (s - c) / s)

    # ------------------------------------------------------------------------------------------------------------------
    def circumcenter(self) -> Vec3:
        a = (self.c.position - self.b.position).length()
        b = (self.c.position - self.a.position).length()
        c = (self.b.position - self.a.position).length()
        q = a * a * (-a * a + b * b + c * c)
        w = b * b * (a * a - b * b + c * c)
        e = c * c * (a * a + b * b - c * c)
        n = q + w + e
        if n == 0.0:
            return Vec3(0, 0, 0)
        return self.fromBarycentric(Vec3(q / n, w / n, e / n))

    # ------------------------------------------------------------------------------------------------------------------
    def incenter(self) -> Vec3:
        a = (self.c.position - self.b.position).length()
        b = (self.c.position - self.a.position).length()
        c = (self.b.position - self.a.position).length()
        n = a + b + c
        if n == 0.0:
            return Vec3(0, 0, 0)

        return self.fromBarycentric(Vec3(a / n, b / n, c / n))

    # ------------------------------------------------------------------------------------------------------------------
    def centroid(self) -> Vec3:
        return (self.a.position + self.b.position + self.c.position) / 3.0

    # ------------------------------------------------------------------------------------------------------------------
    def calcTexcoord(self, p: Vec3) -> Vec2:
        """
        Returns texture-coordinate at cartesian position p
        :param p:
        :return: returns
        """
        pb = self.toBarycentric(p)
        u = self.a.texcoord.x * pb.x + self.b.texcoord.x * pb.y + self.c.texcoord.x * pb.z
        v = self.a.texcoord.y * pb.x + self.b.texcoord.y * pb.y + self.c.texcoord.y * pb.z
        return Vec2(u, v)

    # ------------------------------------------------------------------------------------------------------------------

    def hit(self, ray: Ray, hitrecord: HitRecord) -> bool:
        """
        Ray Triangle Intersection
        Original Code from:
        "Practical Analysis of Optimized Ray-Triangle Intersection"
        Tomas Möller
        Department of Computer Engineering, Chalmers University of Technology, Sweden.
        http://fileadmin.cs.lth.se/cs/Personal/Tomas_Akenine-Moller/raytri/

        :param ray: the ray to check hit
        :param tmin: tmin to test intersection
        :param tmax: tmax to test intersection
        :param hitrecord: the hitrecord which is only valid if there is a hit
        :return: True if there is a hit
        """
        t0 = hitrecord.t

        # find vectors for two edges sharing vert0
        edge1 = self.b.position - self.a.position
        edge2 = self.c.position - self.a.position

        # begin calculating determinant - also used to calculate U parameter
        pvec = cross3(ray.direction, edge2)

        # if determinant is near zero, ray lies in plane of triangle
        det = dot3(edge1, pvec)

        if det > self.EPSILON:
            tvec = ray.start - self.a.position
            u = dot3(tvec, pvec)
            if u < 0.0 or u > det:
                return False

            qvec = cross3(tvec, edge1)
            v = dot3(ray.direction, qvec)
            if v < 0.0 or u + v > det:
                return False
        elif det < -self.EPSILON:
            tvec = ray.start - self.a.position
            u = dot3(tvec, pvec)
            if u > 0.0 or u < det:
                return False

            qvec = cross3(tvec, edge1)
            v = dot3(ray.direction, qvec)
            if v > 0.0 or u + v < det:
                return False
        else:
            return False

        inv_det = 1.0 / det
        t = dot3(edge2, qvec) * inv_det
        u *= inv_det
        v *= inv_det

        if t0 is not None and t > t0:
            return False

        if t > 0.0:  # and t<tmax
            hitrecord.t = t
            hitrecord.point = ray.start + t * ray.direction
            hitrecord.normal = cross3(edge1, edge2)
            if self.a.normal is not None and self.b.normal is not None and self.c.normal is not None:
                nU = self.b.normal - self.a.normal
                nV = self.c.normal - self.a.normal
                hitrecord.normal_g = self.a.normal + nU * u + nV * v
            else:
                hitrecord.normal_g = hitrecord.normal

            # Calculate color
            cU = self.b.color - self.a.color
            cV = self.c.color - self.a.color
            hitrecord.color = self.a.color + cU * u + cV * v
            hitrecord.material = self.material

            hitrecord.point = ray.start + t * ray.direction

            return True

        return False

    def hitShadow(self, ray: Ray) -> bool:
        """
        :param ray:
        :param tmin:
        :param tmax:
        :return:
        """
        # find vectors for two edges sharing vert0
        edge1 = self.b.position - self.a.position
        edge2 = self.c.position - self.a.position

        # begin calculating determinant - also used to calculate U parameter
        pvec = cross3(ray.direction, edge2)

        # if determinant is near zero, ray lies in plane of triangle
        det = dot3(edge1, pvec)

        if det > self.EPSILON:
            tvec = ray.start - self.a.position
            u = dot3(tvec, pvec)
            if u < 0.0 or u > det:
                return False

            qvec = cross3(tvec, edge1)
            v = dot3(ray.direction, qvec)
            if v < 0.0 or u + v > det:
                return False
        elif det < -self.EPSILON:
            tvec = ray.start - self.a.position
            u = dot3(tvec, pvec)
            if u > 0.0 or u < det:
                return False

            qvec = cross3(tvec, edge1)
            v = dot3(ray.direction, qvec)
            if v > 0.0 or u + v < det:
                return False
        else:
            return False

        inv_det = 1.0 / det
        t = dot3(edge2, qvec) * inv_det
        u *= inv_det
        v *= inv_det

        if t > 0.0:
            return True

        return False


# ----------------------------------------------------------------------------------------------------------------------
# Utility functions related to triangles

def triangleArea(a: Vec3, b: Vec3, c: Vec3) -> float:
    """
    Calculate area of triangle
    :return: area
    """
    return cross3(b - a, c - a).length() / 2.0



"""
Vec3: 3 dimensional vector class

This is the class for handling 3 dimensional vectors
"""

import math
from .constants import *


class Vec3(object):
    """Class representing a 3D-Vector. Values are always stored as float
    """

    def __init__(self, *args, **kwargs):
        """
        Vec3 can be constructed the following way:

        v1 = Vec3(x,y,z)    # 3 integers or floats
        v2 = Vec3([x,y,z])  # from a list of 3 integers or floats
        v3 = Vec3((x,y,z))  # from a tuple of 3 integers or floats
        v4 = Vec3(x)

        # using named arguments
        v5 = Vec3(z=1)
        v6 = Vec3(x=1, y=2, z=3)
        v7 = Vec3(y=4, x=2)

        # using nothing (Null-Vector)
        v7 = Vec3()

        """

        if len(kwargs) == 0:
            if len(args) == 3:
                self.x = float(args[0])
                self.y = float(args[1])
                self.z = float(args[2])
            elif len(args) == 1:
                if type(args) == tuple or type(args) == list:
                    self.x = float(args[0][0])
                    self.y = float(args[0][1])
                    self.z = float(args[0][2])
                else:
                    raise ValueError("Sequencial definition of Vec3 must be list or tuple!")

            elif len(args) == 0:
                self.x = self.y = self.z = 0.0

            else:
                raise ValueError("Wrong number of parameters in Vec3")
        else:

            if len(args) > 0:
                raise ValueError("Use keyword arguments OR normal arguments, both is not permitted")

            if "x" in kwargs:
                self.x = float(kwargs["x"])
            else:
                self.x = 0

            if "y" in kwargs:
                self.y = float(kwargs["y"])
            else:
                self.y = 0

            if "z" in kwargs:
                self.z = float(kwargs["z"])
            else:
                self.z = 0

    def __str__(self):
        """Convert vector to string
        """
        return "Vec3(" + str(self.x) + ", " + str(self.y) + ", " + str(self.z) + ")"

    def __add__(self, other):
        """add two vectors
        other: vector to add"""
        return Vec3(self.x + other.x, self.y + other.y, self.z + other.z)

    def __sub__(self, other):
        """subtract two vectors
        other: vector to subtract"""
        return Vec3(self.x - other.x, self.y - other.y, self.z - other.z)

    def __eq__(self, other):
        if type(other) == list or type(other) == tuple:  # not checking type within tuple/list
            if len(other) == 3:
                return (abs(self.x - other[0]) < G_EPSILON) and (abs(self.y - other[1]) < G_EPSILON) and (
                    abs(self.z - other[2]) < G_EPSILON)
            else:
                raise ValueError("Can't compare vector with list or tuple of length != 3")
        elif type(other) == Vec3:
            return (abs(self.x - other.x) < G_EPSILON) and (abs(self.y - other.y) < G_EPSILON) and (
                abs(self.z - other.z) < G_EPSILON)
        else:
            raise ValueError("Can't compare this type")

    def __mul__(self, s):
        if type(s) == Vec3:
            return Vec3(self.x * s.x, self.y * s.y, self.z * s.z)
        elif type(s) == float or type(s) == int:
            return Vec3(self.x * s, self.y * s, self.z * s)
        else:
            raise ValueError("Multiplicaton with wrong type: " + str(type(s)))

    def __rmul__(self, s):
        if type(s) == Vec3:
            return Vec3(self.x * s.x, self.y * s.y, self.z * s.z)
        elif type(s) == float or type(s) == int:
            return Vec3(self.x * s, self.y * s, self.z * s)
        else:
            raise ValueError("Multiplicaton with wrong type:" + str(type(s)))

    def __truediv__(self, s):
        if type(s) == Vec3:
            return Vec3(self.x / s.x, self.y / s.y, self.z / s.z)
        elif type(s) == float or type(s) == int:
            return Vec3(self.x / s, self.y / s, self.z / s)
        else:
            raise ValueError("Division with wrong type" + str(type(s)))

    def __rtruediv__(self, s):
        if type(s) == Vec3:
            return Vec3(s.x / self.x, s.y / self.y, s.z / self.z)
        elif type(s) == float or type(s) == int:
            return Vec3(s / self.x, s / self.y, s / self.z)
        else:
            raise ValueError("Division with wrong type: " + str(type(s)))

    def __neg__(self):
        return Vec3(-self.x, -self.y, -self.z)

    def __getitem__(self, key):
        if key == 0:
            return self.x
        elif key == 1:
            return self.y
        elif key == 2:
            return self.z
        else:
            raise IndexError("Vec3 has 3 components: [0] [1] and [2]")

    def __setitem__(self, key, value):
        if key == 0:
            self.x = float(value)
        elif key == 1:
            self.y = float(value)
        elif key == 2:
            self.z = float(value)
        else:
            raise IndexError("Vec3 has 3 components: [0] [1] and [2]")

    def copy(self):
        """
        :return: copy of current vector
        """
        return Vec3(self.x, self.y, self.z)

    def isZero(self):
        """
        :return: True if this is a zero-vector
        """
        return math.fabs(self.x) < G_EPSILON and math.fabs(self.y) < G_EPSILON and math.fabs(self.z) < G_EPSILON

    def length(self):
        """
        :return: return length of vector
        """
        return math.sqrt(self.x ** 2 + self.y ** 2 + self.z ** 2)

    def normalize(self):
        """
        normalize this vector
        """
        l = self.length()
        if l != 0:
            self.x /= l
            self.y /= l
            self.z /= l



"""
This is the abstract Renderer.

If you override if you implement your own.
"""

from abc import abstractmethod
from ..scene import Scene


class Renderer(object):
    def __init__(self, name="UNKNOWN RENDERER"):
        self.name = name
        print("# Creating Renderer: " + self.name)

    @abstractmethod
    def render(self, scene : Scene) -> list:
        """
        Abstract rendering. This actually renmders the scene and returns the image data

        :param scene: the scene to render
        :return:
        """
        pass



"""
Vec4: 4 dimensional vector class

This is the class for handling 4 dimensional vectors
"""

import math
from .constants import *
from .vec3 import *


class Vec4(object):
    """Class representing a 4D-Vector. Values are always stored as float
    """

    def __init__(self, *args, **kwargs):
        """
        Vec3 can be constructed the following way:

        v1 = Vec4(x,y,z,w)     # 4 integers or floats
        v2 = Vec4([x,y,z,w])   # from a list of 4 integers or floats
        v3 = Vec4((x,y,z,w))   # from a tuple of 4 integers or floats
        v4 = Vec4(x)
        v5 = Vec3(Vec3(1,2,3), 4) # from a Vec3 with optional w-Value (if not set, w=1)

        # using named arguments
        v6 = Vec4(z=1)
        v7 = Vec4(x=1, y=2, z=3)
        v8 = Vec4(y=4, x=2)
        v9 = Vec4(x=2, w=2)

        # using nothing (Null-Vector)
        v10 = Vec4()

        """

        if len(kwargs) == 0:
            if len(args) == 4:
                self.x = float(args[0])
                self.y = float(args[1])
                self.z = float(args[2])
                self.w = float(args[3])
            elif len(args) == 1:
                if type(args) == tuple or type(args) == list:
                    self.x = float(args[0][0])
                    self.y = float(args[0][1])
                    self.z = float(args[0][2])
                    self.w = float(args[0][3])
                elif type(args[0]) == Vec3:
                    self.x = args[0].x
                    self.y = args[0].y
                    self.z = args[0].z
                    self.w = 1.0
                else:
                    raise ValueError("Sequencial definition of Vec4 must be list or tuple!")
            elif len(args) == 2:
                if type(args[0]) == Vec3:
                    self.x = args[0].x
                    self.y = args[0].y
                    self.z = args[0].z
                    self.w = float(args[1])

            elif len(args) == 0:
                self.x = self.y = self.z = 0.0
                self.w = 1.0
            else:
                raise ValueError("Wrong number of parameters in Vec3")
        else:
            if len(args) > 0:
                raise ValueError("Use keyword arguments OR normal arguments, both is not permitted")

            if "x" in kwargs:
                self.x = float(kwargs["x"])
            else:
                self.x = 0

            if "y" in kwargs:
                self.y = float(kwargs["y"])
            else:
                self.y = 0

            if "z" in kwargs:
                self.z = float(kwargs["z"])
            else:
                self.z = 0

            if "w" in kwargs:
                self.w = float(kwargs["w"])
            else:
                self.w = 1.0

    def __str__(self):
        """Convert vector to string
        """
        return "Vec4(" + str(self.x) + ", " + str(self.y) + ", " + str(self.z) + ", " + str(self.w) + ")"

    def __add__(self, other):
        """add two vectors
        other: vector to add"""
        return Vec4(self.x + other.x, self.y + other.y, self.z + other.z, self.w + other.w)

    def __sub__(self, other):
        """subtract two vectors
        other: vector to subtract"""
        return Vec4(self.x - other.x, self.y - other.y, self.z - other.z, self.w - other.w)

    def __mul__(self, s):
        if type(s) == Vec4:
            return Vec4(self.x * s.x, self.y * s.y, self.z * s.z, self.w * s.w)
        elif type(s) == float or type(s) == int:
            return Vec4(self.x * s, self.y * s, self.z * s, self.w * s)
        else:
            raise ValueError("Vec4: Multiplicaton with wrong type: " + str(type(s)))

    def __eq__(self, other):
        if type(other) == list or type(other) == tuple:  # not checking type within tuple/list
            if len(other) == 4:
                return (abs(self.x - other[0]) < G_EPSILON) and \
                       (abs(self.y - other[1]) < G_EPSILON) and \
                       (abs(self.z - other[2]) < G_EPSILON) and \
                       (abs(self.w - other[3]) < G_EPSILON)
            else:
                raise ValueError("Can't compare vector with list or tuple of length != 4")
        elif type(other) == Vec4:
            return (abs(self.x - other.x) < G_EPSILON) and \
                   (abs(self.y - other.y) < G_EPSILON) and \
                   (abs(self.z - other.z) < G_EPSILON) and \
                   (abs(self.w - other.w) < G_EPSILON)
        else:
            raise ValueError("Can't compare this type")

    def __getitem__(self, key):
        if key == 0:
            return self.x
        elif key == 1:
            return self.y
        elif key == 2:
            return self.z
        elif key == 3:
            return self.w
        else:
            raise IndexError("Vec4 has 4 components: [0] [1] [2] and [3]")

    def __setitem__(self, key, value):
        if key == 0:
            self.x = float(value)
        elif key == 1:
            self.y = float(value)
        elif key == 2:
            self.z = float(value)
        elif key == 3:
            self.w = float(value)
        else:
            raise IndexError("Vec4 has 4 components: [0] [1] [2] and [3]")

    def copy(self):
        """
        :return: copy of current vector
        """
        return Vec4(self.x, self.y, self.z, self.w)

    def isZero(self):
        """
        :return: True if this is a zero-vector
        """
        return math.fabs(self.x) < G_EPSILON and \
               math.fabs(self.y) < G_EPSILON and \
               math.fabs(self.z) < G_EPSILON and \
               math.fabs(self.w) < G_EPSILON

    def length(self):
        """
        :return: return length of vector
        """
        return math.sqrt(self.x ** 2 + self.y ** 2 + self.z ** 2 + self.w ** 2)

    def normalize(self):
        """
        normalize this vector
        """
        l = self.length()
        if l != 0:
            self.x /= l
            self.y /= l
            self.z /= l
            self.w /= l



"""
Vector operations

varius utility functions for vectors
"""

import math
from .vec2 import *
from .vec3 import *
from .vec4 import *


# ---------------------------------------------------------------------------------------------------------------------
def cross3(v1 ,v2):
    """
    Cross product for Vec3
    :param v1: first vector
    :param v2: second vector
    :return: cross product
    """
    x = v1.y * v2.z - v1.z * v2.y
    y = v1.z * v2.x - v1.x * v2.z
    z = v1.x * v2.y - v1.y * v2.x
    return Vec3(x, y, z)

# ---------------------------------------------------------------------------------------------------------------------

def cross(v1, v2):
    """Calculates cross product. Consider using cross2, cross3, cross4
    v1: vector
    v2: vector
    """
    if type(v1) == Vec2 and type(v2) == Vec2:
        raise TypeError("Vec2 currently not supported for cross(v1,v2)")
    elif type(v1) == Vec3 and type(v2) == Vec3:
        return cross3(v1,v2)
    elif type(v1) == Vec4 and type(v2) == Vec4:
        raise TypeError("Vec4 currently not supported for cross(v1,v2)")
    else:
        raise TypeError("Wrong type for cross(v1,v2)")

# ---------------------------------------------------------------------------------------------------------------------

def dot3(a,b):
    """calculates the dot product"""
    return a.x * b.x + a.y * b.y + a.z * b.z

# ---------------------------------------------------------------------------------------------------------------------

def normalize3(v):
    """Returns the normalized vector without modifying the original vector"""
    l = v.length()
    if l != 0:
        return Vec3(v.x / l, v.y / l, v.z / l)
    else:
        return Vec3(0.0, 0.0, 0.0)

# ---------------------------------------------------------------------------------------------------------------------

def normalize(v):
    """
    Normalizes a vector, consider using normalize2, normalize3, normalize4
    :param v:
    :return:
    """
    if type(v) == Vec2 :
        raise TypeError("Vec2 currently not supported for normalize(v)")
    elif type(v) == Vec3:
        return normalize3(v)
    elif type(v) == Vec4:
        raise TypeError("Vec4 currently not supported for normalize(v)")
    else:
        raise TypeError("Wrong type for normalize(v)")


# ---------------------------------------------------------------------------------------------------------------------

def reflect3(N, I):
    """Reflects a vector
    N: Normal
    I: Incdient vector"""
    return I + N * (-2.0 * dot3(N, I))

# ---------------------------------------------------------------------------------------------------------------------

def faceforward3(N, I):
    """
    faceforward operation
    N: Normal vector
    I: Incident vector
    """
    if dot3(N, I) < 0:
        return N
    else:
        return -N

# ---------------------------------------------------------------------------------------------------------------------

def refract3(N, I, eta):
    """
    refract operation
    N: Normal vector
    I: Incident vector
    eta: Refraction koefficient
    """
    d = dot3(I, N)
    k = 1 - eta * eta * (1 - d * d)
    if k < 0:
        return I
    else:
        return I * eta - N * (eta * d + math.sqrt(k))

# ---------------------------------------------------------------------------------------------------------------------

def sign3(v):
    def fsign(f):
        if f < 0:
            return -1
        if f > 0:
            return 1
        else:
            return 0

    return Vec3(fsign(v.x), fsign(v.y), fsign(v.z))

# ---------------------------------------------------------------------------------------------------------------------
def clamp(v: float, minVal: float, maxVal: float):
    """
        Component-wise clamp

        In case v is smaller than minVal, minVal is returned.
        If v is larger than maxVal, maxVal is returned.

        :param v: vector to clamp
        :param minVal: minimal value (component-wise)
        :param maxVal: maximal value (component-wise)
        :return: clamped value
    """
    if v<minVal:
        return minVal
    elif v>maxVal:
        return maxVal
    else:
        return v

# ---------------------------------------------------------------------------------------------------------------------

def clamp2(v: Vec2, minVal: Vec2, maxVal: Vec2):
    """
    Component-wise clamp

    In case v is smaller than minVal, minVal is returned.
    If v is larger than maxVal, maxVal is returned.

    :param v: vector to clamp
    :param minVal: minimal value (component-wise)
    :param maxVal: maximal value (component-wise)
    :return: clamped value
    """
    result = v.copy()
    if v.x < minVal.x:
        result.x = minVal.x
    if v.y < minVal.y:
        result.y = minVal.y
    if v.x > maxVal.x:
        result.x = maxVal.x
    if v.y > maxVal.y:
        result.y = maxVal.y

    return result

# ---------------------------------------------------------------------------------------------------------------------


def clamp3(v: Vec3, minVal: Vec3, maxVal:Vec3):
    """
    Component-wise clamp

    In case v is smaller than minVal, minVal is returned.
    If v is larger than maxVal, maxVal is returned.

    :param v: vector to clamp
    :param minVal: minimal value (component-wise)
    :param maxVal: maximal value (component-wise)
    :return: clamped value
    """
    result = v.copy()
    if v.x < minVal.x:
        result.x = minVal.x
    if v.y < minVal.y:
        result.y = minVal.y
    if v.z < minVal.z:
        result.z = minVal.z
    if v.x > maxVal.x:
        result.x = maxVal.x
    if v.y > maxVal.y:
        result.y = maxVal.y
    if v.z > maxVal.z:
        result.z = maxVal.z

    return result

# ---------------------------------------------------------------------------------------------------------------------


def clamp4(v: Vec4, minVal: Vec4, maxVal: Vec4):
    """
    Component-wise clamp

    In case v is smaller than minVal, minVal is returned.
    If v is larger than maxVal, maxVal is returned.

    :param v: vector to clamp
    :param minVal: minimal value (component-wise)
    :param maxVal: maximal value (component-wise)
    :return: clamped value
    """
    result = v.copy()
    if v.x < minVal.x:
        result.x = minVal.x
    if v.y < minVal.y:
        result.y = minVal.y
    if v.z < minVal.z:
        result.z = minVal.z
    if v.w < minVal.w:
        result.w = minVal.w
    if v.x > maxVal.x:
        result.x = maxVal.x
    if v.y > maxVal.y:
        result.y = maxVal.y
    if v.z > maxVal.z:
        result.z = maxVal.z
    if v.w > maxVal.w:
        result.w = maxVal.w

    return result




"""
Base Material

This is the description of a base material. If you want to add new materials, inherit from this class.
"""

from abc import abstractmethod
from ..math import Vec3, Ray, HitRecord
from ..camera import Camera


class Material(object):

    """Base Material Class"""

    def __init__(self, color: Vec3 = Vec3(0.,0.,0.), shininess: float = 0.0, reflectivity: float = 0.5, refraction: float = 1.0):
        self.color = color
        self.shininess = shininess
        self.reflectivity = reflectivity
        """
        some important refractive indices:

        vacuum: 1.0
        water: 1.33
        diamond: 2.4
        more indices can be found at: https://en.wikipedia.org/wiki/List_of_refractive_indices

        Also check the "refract3" function in math/vecops.py
        """
        self.refraction = refraction

    @abstractmethod
    def shade(self, camera: Camera, ray: Ray, hitrecord: HitRecord,  lights: list) -> Vec3:
        """
        Shade method

        This is the base shade method.
        """
        pass




"""
This is the definition for a point light

A point light emits light to all directions. It can be somewhat compared to a light bulb.

"""
from .light import Light
from ..math import Vec3


class PointLight(Light):

    """Class describing a point light source"""

    def __init__(self, position: Vec3) -> None:
        """
        Constructor

        Not yet complete - params will change in near future!
        :param position: position of the light
        """
        Light.__init__(self, "PointLight")
        self.position = position






from ..math import Vec2, Vec3


class RGBImage(object):
    def __init__(self, width: int, height: int) -> None:
        self.width = width
        self.height = height
        self.type = "RGB"

        if self.type == "RGB":
            self.data = [(0, 0, 0) for i in range(0, self.width*self.height)]
        else:   # in future there will be RGBA, FLOAT images
            raise ValueError("unknown/unsupported image type")

    #-------------------------------------------------------------------------------------------------------------------

    def drawPoint(self, pos: Vec2, color: Vec3) -> None:
        # This code is really slow if you write many pixels - it even contains a bounds check.
        # If you draw many points, please use self.data directly.
        # This draw point assumes (0,0) to be bottom left
        x = pos.x
        y = self.height-pos.y-1
        if x>=0 and x<self.width and y>=0 and y<self.height:
            self.data[y*self.width+x] = (int(color[0]*255.), int(color[1]*255.), int(color[2]*255.))

    #-------------------------------------------------------------------------------------------------------------------

    def drawLine(self, start: Vec2, end: Vec2, color: Vec3 = Vec3(1.,1.,1.)) -> None:
        """Bresenham's line algorithm
           modified from: https://rosettacode.org/wiki/Bitmap/Bresenham%27s_line_algorithm#Python
        """
        dx = abs(end.x - start.x)
        dy = abs(end.y - start.y)
        x, y = start.x, start.y
        sx = -1 if start.x > end.x else 1
        sy = -1 if start.y > end.y else 1
        if dx > dy:
            err = dx / 2.0
            while x != end.x:
                self.drawPoint(Vec2(x, y), color)
                err -= dy
                if err < 0:
                    y += sy
                    err += dx
                x += sx
        else:
            err = dy / 2.0
            while y != end.y:
                self.drawPoint(Vec2(x, y), color)
                err -= dx
                if err < 0:
                    x += sx
                    err += dy
                y += sy
        self.drawPoint(Vec2(x, y), color)

    #-------------------------------------------------------------------------------------------------------------------

    def drawCircle(self, center: Vec2, radius: int, color: Vec3):
        switch = 3 - (2 * radius)
        x = 0
        y = radius
        while x <= y:
            self.drawPoint(Vec2(x + center.x, -y + center.y), color)
            self.drawPoint(Vec2(y + center.x, -x + center.y), color)
            self.drawPoint(Vec2(y + center.x, x + center.y), color)
            self.drawPoint(Vec2(x + center.x, y + center.y), color)
            self.drawPoint(Vec2(-x + center.x, y + center.y), color)
            self.drawPoint(Vec2(-y + center.x, x + center.y), color)
            self.drawPoint(Vec2(-y + center.x, -x + center.y), color)
            self.drawPoint(Vec2(-x + center.x, -y + center.y), color)
            if switch < 0:
                switch = switch + (4 * x) + 6
            else:
                switch = switch + (4 * (x - y)) + 10
                y = y - 1
            x = x + 1

    #-------------------------------------------------------------------------------------------------------------------

    def drawRectangle(self, bl : Vec2, width: int, height: int, color: Vec3) -> None:
        """
        Draws a rectangle.
        This is quite slow and would be much faster using direct image buffer filling instead of using bresenham lines.

        :param bl: bottom left coordinate of center
        :param width: width of rectangle
        :param height: height of rectangle
        :param color: color of rectangle
        :return:
        """
        for y in range(height):
            start = Vec2(bl.x, bl.y + y)
            end = Vec2(bl.x + width, bl.y + y)
            self.drawLine(start, end, color)







"""
Base class for cameras

Inherit from this if you create a new camera model
"""

from abc import abstractmethod
from ..math import Vec3, createLookAt4, createIdentity4, Ray, Mat4


class Camera(object):

    """This defines the base class for all cameras"""

    def __init__(self):
        self.position = Vec3(0.,0.,0.)
        self.view = createIdentity4()

    def setView(self, eye: Vec3, center: Vec3, up: Vec3):
        """
        Set View Matrix using look-at (from eye to center)

        :param eye: current position
        :param center: position where to look at
        :param up: up-axis when looking at
        :return:
        """
        self.view = createLookAt4(eye, center, up)

    @abstractmethod
    def primaryRay(self, x: float, y: float) -> Ray:
        """
        Calculate primary ray

        :param x: screen position x (subpixel)
        :param y: screen position y (subpixel)
        :return: ray
        """
        pass

    @abstractmethod
    def getMatrix(self) -> Mat4:
        """
        Returns the view-projection matrix

        :return:
        """
        pass



"""
Vec2: 2 dimensional vector class

This is the class for handling 2 dimensional vectors
"""

import math
from .constants import *


class Vec2(object):
    """Class representing a 2D-Vector. Values are always stored as float
    """

    def __init__(self, x, y):
        self.x = x
        self.y = y


    def __str__(self):
        """Convert vector to string
        """
        return "Vec2(" + str(self.x) + ", " + str(self.y) + ")"

    def __add__(self, other):
        """add two vectors
        other: vector to add"""
        return Vec2(self.x + other.x, self.y + other.y)

    def __sub__(self, other):
        """subtract two vectors
        other: vector to subtract"""
        return Vec2(self.x - other.x, self.y - other.y)

    def __eq__(self, other):
        if type(other) == list or type(other) == tuple:  # not checking type within tuple/list
            if len(other) == 2:
                return (abs(self.x - other[0]) < G_EPSILON) and (abs(self.y - other[1]) < G_EPSILON)
            else:
                raise ValueError("Can't compare vector with list or tuple of length != 3")
        elif type(other) == Vec2:
            return (abs(self.x - other.x) < G_EPSILON) and (abs(self.y - other.y) < G_EPSILON)
        else:
            raise ValueError("Can't compare this type")

    def __mul__(self, s):
        if type(s) == Vec2:
            return Vec2(self.x * s.x, self.y * s.y)
        elif type(s) == float or type(s) == int:
            return Vec2(self.x * s, self.y * s)
        else:
            raise ValueError("Multiplicaton with wrong type: " + str(type(s)))

    def __rmul__(self, s):
        if type(s) == Vec2:
            return Vec2(self.x * s.x, self.y * s.y)
        elif type(s) == float or type(s) == int:
            return Vec2(self.x * s, self.y * s)
        else:
            raise ValueError("Multiplicaton with wrong type:" + str(type(s)))

    def __truediv__(self, s):
        if type(s) == Vec2:
            return Vec2(self.x / s.x, self.y / s.y)
        elif type(s) == float or type(s) == int:
            return Vec2(self.x / s, self.y / s)
        else:
            raise ValueError("Division with wrong type" + str(type(s)))

    def __rtruediv__(self, s):
        if type(s) == Vec2:
            return Vec2(s.x / self.x, s.y / self.y)
        elif type(s) == float or type(s) == int:
            return Vec2(s / self.x, s / self.y)
        else:
            raise ValueError("Division with wrong type: " + str(type(s)))

    def __neg__(self):
        return Vec2(-self.x, -self.y)

    def __getitem__(self, key):
        if key == 0:
            return self.x
        elif key == 1:
            return self.y
        else:
            raise IndexError("Vec2 has 3 components: [0] [1] and [2]")

    def __setitem__(self, key, value):
        if key == 0:
            self.x = float(value)
        elif key == 1:
            self.y = float(value)
        else:
            raise IndexError("Vec2 has 3 components: [0] [1] and [2]")

    def copy(self):
        """
        :return: copy of current vector
        """
        return Vec2(self.x, self.y)

    def length(self):
        """
        :return: return length of vector
        """
        return math.sqrt(self.x ** 2 + self.y ** 2)

    def normalize(self):
        """
        normalize this vector
        """
        l = self.length()
        if l != 0:
            self.x /= l
            self.y /= l



"""
Math for pyRT

This contains vector, matrix operations and other mathematical utils like planes, rays, ...
"""

from .constants import G_EPSILON
from .ray import Ray, HitRecord
from .vec3 import Vec3
from .vec4 import Vec4
from .mat4 import Mat4
from .vecops import *
from .matops import *



"""
This module contains different camera specifications.

Especially perspective and orthographic cameras.
"""

from .camera import Camera
from .perspectivecamera import PerspectiveCamera
from .orthographiccamera import OrthographicCamera



"""
A simple ray tracer

Reference implementation...
"""

from .renderer import Renderer
from ..scene import Scene
from ..math import *
import time


class SimpleRT(Renderer):
    def __init__(self, shadow=False, iterations=1):
        Renderer.__init__(self, "Simple Raytracer")
        self.shadow = shadow
        self.iterations = iterations
        if self.shadow:
            print("# Shadow Enabled")
        if self.iterations>1:
            print("# Iterations: " + str(self.iterations))

    def _shade(self, scene: Scene, ray: Ray, hitrecord: HitRecord) -> tuple:
        r = g = b = 0  # background color
        hit = False
        for element in scene.nodes:
            if element.hit(ray, hitrecord):
                hit = True
                hitrecord.obj = element  # set hit object

                # element hit -> call shader:
                color = element.material.shade(scene.camera, ray, hitrecord, scene.lights)
                r = int(color[0] * 255)
                g = int(color[1] * 255)
                b = int(color[2] * 255)

                # Calculate shadow
                if self.shadow:
                    # is hitpoint in shadow ?
                    f, n = self._shadow(scene, hitrecord)
                    self.num_shadow_rays += n
                    r = int(r * f)
                    g = int(g * f)
                    b = int(b * f)


        return hit, r, g, b

    def _shadow(self, scene: Scene, hitrecord: HitRecord) -> tuple:
        # is hitpoint in shadow ?
        fs = 1.0
        local_num_shadow_rays = 0
        numshadow = 0
        for light in scene.lights:
            shadowray = Ray(hitrecord.point, light.position - hitrecord.point)
            local_num_shadow_rays += 1
            for testelement in scene.nodes:
                if testelement != hitrecord.obj:  # avoid self-intersection
                    if testelement.hitShadow(shadowray):
                        numshadow += 1
                        break
        for i in range(numshadow):
            fs /= 4.

        return fs,local_num_shadow_rays

    def _reflect(self, r: int, g: int, b: int, scene: Scene, ray: Ray, hitrecord: HitRecord) -> tuple:
        reflect_ray = Ray(hitrecord.point, reflect3(hitrecord.normal_g, ray.direction))
        reflect_hitrecord = HitRecord()

        hit, rnew, gnew, bnew = self._shade(scene, reflect_ray, reflect_hitrecord)

        if hit:
            ref1 = hitrecord.material.reflectivity
            ref2 = 1.-ref1
            rnew = int ( ref1 * rnew + ref2 * r)
            gnew = int ( ref1 * gnew + ref2 * g)
            bnew = int ( ref1 * bnew + ref2 * b)
            return rnew, gnew, bnew, reflect_ray, reflect_hitrecord
        else:
            return r,g,b,None,None

    def render(self, scene: Scene) -> list:
        if not scene.camera:
            print("Warning: Can't render: there is no (active) camera in the scene!")
            return None

        time_start = time.time()
        self.num_rays = 0
        self.num_shadow_rays  = 0
        self.num_secondary_rays = 0
        hitrecord = HitRecord()

        w = scene.camera.width
        h = scene.camera.height

        image = []
        for y in range(0, h):
            for x in range(0, w):
                ray = scene.camera.primaryRay(x, y)
                hitrecord.reset()
                self.num_rays += 1

                # Primary Ray:
                hit, r, g, b = self._shade(scene, ray, hitrecord)

                if hit and hitrecord.material.reflectivity != 0.0:
                    refhit = hitrecord.copy()
                    refray = ray.copy()
                    for i in range(self.iterations - 1):
                        r, g, b, refray, refhit = self._reflect(r,g,b, scene, refray, refhit)
                        self.num_secondary_rays += 1
                        if refray is None:
                            break

                image.append((r, g, b))

        time_end = time.time()
        print("# RENDER STATISTICS" + 31 * "#")
        time_total = time_end - time_start
        print("TIME FOR RENDERING: " + str(time_total) + "s")
        print("NUMBER OF PRIMARY RAYS: " + str(self.num_rays))
        print("NUMBER OF SECONDARY RAYS: " + str(self.num_secondary_rays))
        print("NUMBER OF SHADOW RAYS: " + str(self.num_shadow_rays))
        print("RAYS/s: " + str((self.num_rays + self.num_secondary_rays + self.num_shadow_rays) / time_total))
        print(50 * "#")

        return image



"""
Here the vertex is defined

The following attributes are supported:
  * position
  * normal
  * color
  * texcoord
"""

from ..math import *


class Vertex(object):
    def __init__(self, **kwargs):
        self.position = Vec3(0, 0, 0)
        self.normal = None
        self.color = Vec3(1.,1.,1.)
        self.texcoord = None

        if len(kwargs) == 0:
            raise ValueError("Vertex must specify attributes, for example: position, normal, color, texcoord")
        else:

            if "position" in kwargs:
                if type(kwargs["position"]) == Vec3:
                    self.position = kwargs["position"].copy()
                elif type(kwargs["position"]) == tuple or type(kwargs["position"]) == list:
                    if len(kwargs["position"]) != 3:
                        raise ValueError("position must be specified as Vec3 or a list/tuple")
                    else:
                        self.position = Vec3(kwargs["position"][0], kwargs["position"][1], kwargs["position"][2])
                else:
                    raise ValueError("position must be specified as Vec3 or a list/tuple")

            if "normal" in kwargs:
                if type(kwargs["normal"]) == Vec3:  # TODO: a list/tuple of values is ok too...
                    self.normal = kwargs["normal"].copy()
                elif type(kwargs["normal"]) == tuple or type(kwargs["normal"]) == list:
                    if len(kwargs["normal"]) != 3:
                        raise ValueError("normal must be specified as Vec3 or a list/tuple with 3 floats")
                    else:
                        self.normal = Vec3(kwargs["position"][0], kwargs["position"][1], kwargs["position"][2])
                else:
                    raise ValueError("normal must be specified as Vec3 or a list/tuple")

            if "color" in kwargs:
                if type(kwargs["color"]) == Vec4:
                    raise ValueError("Color must be Vec3")
                elif type(kwargs["color"]) == Vec3:
                    color = kwargs["color"]
                    self.color = Vec3(color.x, color.y, color.z)
                elif type(kwargs["color"]) == tuple or type(kwargs["color"]) == list:
                    color = kwargs["color"]
                    if len(color) == 3:
                        self.color = Vec3(color[0], color[1], color[2])
                    else:
                        raise ValueError("Wrong number of components for color")
                else:
                    raise ValueError("color must be specified as Vec3 or list/tuple with 3 components")

            '''
            if "texcoord" in kwargs:
                if type(kwargs["texcoord"]) == Vec2:
                    pass
                elif type(kwargs["texcoord"]) == tuple or type(kwargs["texcoord"]) == list:
                else:
                    raise ValueError("position must be specified as Vec2 or list/tuple")
            '''



"""
Orthographic Camera

This is for a standard orthographic camera
"""

from ..math import Ray, Vec3
from .camera import Camera

class OrthographicCamera(Camera):

    """This defines an orthographic camera"""

    def __init__(self, width=512, height=512):
        """
        :param width: horizontal resolution of the output image
        :param height: vertical resolution of the output image
        """
        Camera.__init__(self)

        self.width = width
        self.height = height
        self.direction = Vec3(0.0, 0.0, -1.0)

        print("[init] Orthographic Camera")

    def primaryRay(self, x: float, y: float) -> Ray:
        r = Ray(Vec3(x, y, 0.), self.direction)
        return r



"""
Phong Material

For phong shading
"""
from .material import Material
from ..math import Vec3, Ray, HitRecord, dot3, reflect3, normalize3, clamp3
from ..camera import Camera


class PhongMaterial(Material):

    """Base Material Class"""

    def __init__(self, color: Vec3 = Vec3(1.,1.,1.), shininess: float = 10.0, reflectivity: float = 0.0, refraction: float = 1.0):
        Material.__init__(self, color, shininess, reflectivity, refraction)

    def shade(self, camera: Camera, ray: Ray, hitrecord: HitRecord,  lights: list) -> Vec3:
        """
        Shade method: Phong

        phong shader
        """
        colorsum = Vec3(0.,0.,0.)

        if len(lights)>0:
            for light in lights:
                N = hitrecord.normal_g
                L = normalize3(hitrecord.point - light.position)
                E = normalize3(camera.position - hitrecord.point)
                R = normalize3(-reflect3(L, N))
                diffuse = max(1. - dot3(N, L), 0.0)
                specular = pow(max(dot3(R, E), 0.0), 0.3 * self.shininess)
                color = self.color * 0.5 * (diffuse + specular) * hitrecord.color
                colorsum += color
            colorsum /= len(lights)
            colorsum = clamp3(colorsum, Vec3(0.,0.,0.), Vec3(1.,1.,1.))
        else:
            # no light in scene, use material color
            colorsum = self.color * hitrecord.color

        return colorsum




"""
Mat4: 4x4 Matrix Class

This is the class for handling 4x4 Matrices
"""


import math
from .constants import *
from .vec3 import *
from .vec4 import *


class Mat4(object):

    """
    Class representing a 4x4 Matrix

    It contains all important methods for matrices. Some external operations are defined in "matops.py".
    """

    def __init__(self, elm=None):
        """
        elm:
            if None this will create a matrix initialized with 0

            if elm is a list or a tuple with 16 values, it will be initialized as 4x4 matrix
        """
        if isinstance(elm, type(None)):
            self.m = [0., 0., 0., 0.,
                      0., 0., 0., 0.,
                      0., 0., 0., 0.,
                      0., 0., 0., 0.]
        elif type(elm) == list or type(elm) == tuple:
            if len(elm) == 16:
                self.m = []
                for v in elm:
                    self.m.append(float(v))
            else:
                raise ValueError("Mat4 must be initialized in a list or tuple with 16 floats")

    def __str__(self):
        """Convert to string"""
        s = "[[" + str(self.m[0]) + ", " + str(self.m[1]) + ", " + str(self.m[2]) + ", " + str(self.m[3]) + "]\n"
        s += "[" + str(self.m[4]) + ", " + str(self.m[5]) + ", " + str(self.m[6]) + ", " + str(self.m[7]) + "]\n"
        s += "[" + str(self.m[8]) + ", " + str(self.m[9]) + ", " + str(self.m[10]) + ", " + str(self.m[11]) + "]\n"
        s += "[" + str(self.m[12]) + ", " + str(self.m[13]) + ", " + str(self.m[14]) + ", " + str(self.m[15]) + "]]"
        return s

    def __getitem__(self, key):
        if type(key) == tuple:
            if len(key) != 2:
                raise IndexError("Index must be 2-dimensional!")
            x, y = key

            if x < 0 or x > 4 or y < 0 or y > 4:
                raise IndexError("Index out of range!")

            return self.m[x + y * 4]
        else:
            raise IndexError("Matrix indices must be specified as tuple, for example:   s = m[1,2]")

    def __setitem__(self, key, value):
        if type(key) == tuple:
            if len(key) != 2:
                raise IndexError("Index must be 2-dimensional!")
            x, y = key

            if x < 0 or x > 4 or y < 0 or y > 4:
                raise IndexError("Index out of range!")

            self.m[x + y * 4] = value

        else:
            raise IndexError("Matrix indices must be access as tuple, for example:   m[1,2] = 5")

    def transpose(self):
        """Transpose matrix"""
        t = self.m.copy()
        for x in range(0, 4):
            for y in range(0, 4):
                index0 = x + y * 4
                index1 = y + x * 4
                self.m[index0] = t[index1]

    def __eq__(self, other):
        if type(other) == list or type(other) == tuple:  # not checking type within tuple/list
            if len(other) == 16:
                for i in range(0, 16):
                    if abs(self.m[i] - other[i]) > G_EPSILON:
                        return False
                return True
            else:
                raise ValueError("Can't compare 4x4 Matrix with list or tuple of length != 16")
        elif type(other) == Mat4:
            for i in range(0, 16):
                if (abs(self.m[i] - other.m[i]) > G_EPSILON):
                    return False
            return True
        else:
            raise ValueError("Can't compare matrix with " + str(type(other)))

    def __add__(self, other):
        """add two Mat4"""
        if type(other) == Mat4:
            return Mat4((self.m[0] + other.m[0], self.m[1] + other.m[1], self.m[2] + other.m[2], self.m[3] + other.m[3],
                         self.m[4] + other.m[4], self.m[5] + other.m[5], self.m[6] + other.m[6], self.m[7] + other.m[7],
                         self.m[8] + other.m[8], self.m[9] + other.m[9], self.m[10] + other.m[10],
                         self.m[11] + other.m[11],
                         self.m[12] + other.m[12], self.m[13] + other.m[13], self.m[14] + other.m[14],
                         self.m[15] + other.m[15]))
        else:
            raise ValueError("Wrong type for matrix addition: " + str(type(other)))

    def __sub__(self, other):
        """subtract two Mat4"""
        if type(other) == Mat4:
            return Mat4((self.m[0] - other.m[0], self.m[1] - other.m[1], self.m[2] - other.m[2], self.m[3] - other.m[3],
                         self.m[4] - other.m[4], self.m[5] - other.m[5], self.m[6] - other.m[6], self.m[7] - other.m[7],
                         self.m[8] - other.m[8], self.m[9] - other.m[9], self.m[10] - other.m[10],
                         self.m[11] - other.m[11],
                         self.m[12] - other.m[12], self.m[13] - other.m[13], self.m[14] - other.m[14],
                         self.m[15] - other.m[15]))
        else:
            raise ValueError("Wrong type for matrix subtraction: " + str(type(other)))

    def __mul__(self, other):
        """
        The following multiplications are supported:

        Matrix4 Matrix4 multiplication
        Matrix4 Vector3 multiplication
        Matrix4 Vector4 multiplication
        """
        if type(other) == Mat4:
            newmat = Mat4()
            newmat.m[0] = other.m[0] * self.m[0]\
                          + other.m[4] * self.m[1] \
                          + other.m[8] * self.m[2] \
                          + other.m[12] * self.m[3]
            newmat.m[4] = other.m[0] * self.m[4] \
                          + other.m[4] * self.m[5] \
                          + other.m[8] * self.m[6] \
                          + other.m[12] * self.m[7]
            newmat.m[8] = other.m[0] * self.m[8] \
                          + other.m[4] * self.m[9] \
                          + other.m[8] * self.m[10] \
                          + other.m[12] * self.m[11]
            newmat.m[12] = other.m[0] * self.m[12] \
                           + other.m[4] * self.m[13] \
                           + other.m[8] * self.m[14] \
                           + other.m[12] * self.m[15]
            newmat.m[1] = other.m[1] * self.m[0] \
                          + other.m[5] * self.m[1] \
                          + other.m[9] * self.m[2] \
                          + other.m[13] * self.m[3]
            newmat.m[5] = other.m[1] * self.m[4] \
                          + other.m[5] * self.m[5] \
                          + other.m[9] * self.m[6] \
                          + other.m[13] * self.m[7]
            newmat.m[9] = other.m[1] * self.m[8] \
                          + other.m[5] * self.m[9] \
                          + other.m[9] * self.m[10] \
                          + other.m[13] * self.m[11]
            newmat.m[13] = other.m[1] * self.m[12] \
                           + other.m[5] * self.m[13] \
                           + other.m[9] * self.m[14] \
                           + other.m[13] * self.m[15]
            newmat.m[2] = other.m[2] * self.m[0] \
                          + other.m[6] * self.m[1] \
                          + other.m[10] * self.m[2] \
                          + other.m[14] * self.m[3]
            newmat.m[6] = other.m[2] * self.m[4] \
                          + other.m[6] * self.m[5] \
                          + other.m[10] * self.m[6] \
                          + other.m[14] * self.m[7]
            newmat.m[10] = other.m[2] * self.m[8] \
                           + other.m[6] * self.m[9] \
                           + other.m[10] * self.m[10] \
                           + other.m[14] * self.m[11]
            newmat.m[14] = other.m[2] * self.m[12] \
                           + other.m[6] * self.m[13] \
                           + other.m[10] * self.m[14] \
                           + other.m[14] * self.m[15]
            newmat.m[3] = other.m[3] * self.m[0] \
                          + other.m[7] * self.m[1] \
                          + other.m[11] * self.m[2] \
                          + other.m[15] * self.m[3]
            newmat.m[7] = other.m[3] * self.m[4] \
                          + other.m[7] * self.m[5] \
                          + other.m[11] * self.m[6] \
                          + other.m[15] * self.m[7]
            newmat.m[11] = other.m[3] * self.m[8] \
                           + other.m[7] * self.m[9] \
                           + other.m[11] * self.m[10] \
                           + other.m[15] * self.m[11]
            newmat.m[15] = other.m[3] * self.m[12] \
                           + other.m[7] * self.m[13] \
                           + other.m[11] * self.m[14] \
                           + other.m[15] * self.m[15]
            return newmat

        if type(other) == Vec3:
            result = Vec3()
            result.x = self.m[0] * other.x + self.m[1] * other.y + self.m[2] * other.z + self.m[3]
            result.y = self.m[4] * other.x + self.m[5] * other.y + self.m[6] * other.z + self.m[7]
            result.z = self.m[8] * other.x + self.m[9] * other.y + self.m[10] * other.z + self.m[11]
            w = self.m[12] * other.x + self.m[13] * other.y + self.m[14] * other.z + self.m[15]

            result.x = result.x / w
            result.y = result.y / w
            result.z = result.z / w

            return result

        if type(other) == Vec4:
            result = Vec4()
            result.x = self.m[0] * other.x + self.m[1] * other.y + self.m[2] * other.z + self.m[3]
            result.y = self.m[4] * other.x + self.m[5] * other.y + self.m[6] * other.z + self.m[7]
            result.z = self.m[8] * other.x + self.m[9] * other.y + self.m[10] * other.z + self.m[11]
            result.w = self.m[12] * other.x + self.m[13] * other.y + self.m[14] * other.z + self.m[15]

            return result
        else:
            raise ValueError("Can't multiply matrix with specified type: " + str(type(other)))



"""
This is the abstract Light.

Override if you implement your own light.
"""

#from abc import abstractmethod
from ..math import Vec3
from ..scene import Scene


class Light(object):

    """Base Light Class"""

    def __init__(self, name="UNKNOWN LIGHT"):
        """
        Base Constructor

        :param name: Name of Light type, for example "PointLight".
        """
        self.name = name
        self.position = Vec3(0.,0.,0.)



"""
This is the geometric object sphere

(renderable object)
"""

from ..geometry import Shape
from ..math import Ray, HitRecord, Vec3, dot3, G_EPSILON
from ..material import Material, PhongMaterial
from math import sqrt


class Sphere(Shape):

    """The Sphere class for raytracing"""

    def __init__(self, center: Vec3, radius: float, material: Material = PhongMaterial()) -> None:
        Shape.__init__(self, "Sphere")

        self.center = center
        self.radius = radius
        self.material = material


    def hit(self, ray: Ray, hitrecord: HitRecord) -> bool:
        """
        Hit ray with sphere.

        :param ray: the ray to check hit
        :param hitrecord: the hitrecord which is only valid if there is a hit
        :return: True if there is a hit
        """
        t0 = hitrecord.t

        a = dot3(ray.direction, ray.direction)
        b = dot3(ray.direction, (2.0 * (ray.start - self.center)))
        c = dot3(self.center, self.center) + dot3(ray.start, ray.start) \
            - 2.0 * dot3(ray.start,self.center) - self.radius * self.radius
        D = b * b + (-4.0) * a * c

        if D < G_EPSILON:
            return False

        D = sqrt(D)
        t = -0.5*(b + D) / a

        if t0 is not None and t0 < t:
            return False

        if t>0:
            hitrecord.t = t
            hitrecord.point = ray.start + t * ray.direction
            hitrecord.normal_g = (hitrecord.point - self.center) / self.radius
            hitrecord.normal = hitrecord.normal_g
            hitrecord.color = Vec3(1., 1., 1.)  # spheres don't have interpolated colors, set to white
            hitrecord.material = self.material
            return True
        return False



    def hitShadow(self, ray: Ray) -> bool:
        """
        :param ray:
        :param tmin:
        :param tmax:
        :return:
        """
        a = dot3(ray.direction, ray.direction)
        b = dot3(ray.direction, (2.0 * (ray.start - self.center)))
        c = dot3(self.center, self.center) + dot3(ray.start, ray.start) \
            - 2.0 * dot3(ray.start, self.center) - self.radius * self.radius
        D = b * b + (-4.0) * a * c

        if D < G_EPSILON:
            return False

        D = sqrt(D)
        t = -0.5 * (b + D) / a

        if t > 0:
            return True
        return False


"""
Perspective Camera

This is for a standard perspective camera
"""

from ..math import Vec3, Mat4, Ray, createPerspective4
from ..math.matops import createLookAt4, inverse4
from .camera import Camera


class PerspectiveCamera(Camera):

    """This defines a perspective camera"""

    def __init__(self, width: int = 640, height: int = 480, fov: float = 45) -> None:
        """
        Initialize perspective camera

        :param width: horizontal resolution of the output image
        :param height: vertical resolution of the output image
        :param fov: field of view (degree)
        """
        Camera.__init__(self)

        self.width = width
        self.height = height
        znear = 0.1
        zfar = 1000.0
        aspect = width / height
        self.projection = createPerspective4(fov, aspect, znear, zfar)

        # initialize with a view matrix
        self.setView(Vec3(0, -10, 0), Vec3(0, 0, 0), Vec3(0, 0, 1))

    def primaryRay(self, x: float, y: float) -> Ray:
        planepoint = self.pos3 + self.ddx2 + self.ddy2 + x*self.ddx - y*self.ddy
        direction = (planepoint - self.position)
        return Ray(self.position, direction)



    def setView(self, eye: Vec3, center: Vec3, up: Vec3):
        """
        Set View Matrix using look-at (from eye to center)

        :param eye: current position
        :param center: position where to look at
        :param up: up-axis when looking at
        :return:
        """
        self.view = createLookAt4(eye, center, up)

        self.matrix = self.projection * self.view
        self.matrixinv = inverse4(self.matrix)
        self.viewinv = inverse4(self.view)
        self.position = self.viewinv * Vec3(0, 0, 0)

        # Extract image plane points in world coordinates:
        self.pos0 = self.matrixinv * Vec3(-1, -1, -1)
        self.pos1 = self.matrixinv * Vec3(1, -1, -1)
        self.pos2 = self.matrixinv * Vec3(1, 1, -1)
        self.pos3 = self.matrixinv * Vec3(-1, 1, -1)

        self.dx = self.pos2-self.pos3   # vector along x-axis of image plane
        self.dy = self.pos3-self.pos0   # vector along y-axis of image plane

        self.ddx = self.dx/self.width   # vector for one pixel (x-axis)
        self.ddy = self.dy/self.height  # vector for one pixel (y-axis)
        self.ddx2 = self.ddx / 2. # half pixel x
        self.ddy2 = self.ddy / 2. # half pixel y

    # pylint: disable-msg=R0913
    def setProjection(self, fov: float, width: int, height: int, znear: float, zfar: float):
        """
        Redefine perspective projection.

        :param fov: Field of view
        :param width: width in pixels
        :param height: height in pixels
        :param znear: near-plane
        :param zfar: far-plane
        :return:
        """
        self.width = width
        self.height = height
        aspect = self.width / self.height
        self.projection = createPerspective4(fov, aspect, znear, zfar)
        self.matrix = self.projection * self.view
        self.matrixinv = inverse4(self.matrix)


    def getMatrix(self) -> Mat4:
        """
        Returns the view-projection matrix

        :return: view-projection
        """
        return self.matrix




"""
This is the geometry module, containing maily shapes.

Also the vertex specification is found here.
"""

from .vertex import Vertex
from .shape import Shape
from .triangle import Triangle
from .sphere import Sphere



"""
pyRT - The Python Raytracer

This is the Setup-Script...
"""

#calls for installation:
#    python3 setup.py install
#
#calls for distribution:
#    set HOME=C:\users\username    # or wherever your .pypirc is
#    python3 setup.py sdist
#    python3 setup.py sdist upload


from setuptools import setup
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
# README.rst was created using: run pandoc --from=markdown --to=rst --output=README.rst README.md
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='pyrt',
    packages=['pyrt',
              'pyrt.camera',
              'pyrt.geometry',
              'pyrt.light',
              'pyrt.material',
              'pyrt.math',
              'pyrt.renderer',
              'pyrt.scene'],
    package_dir={
        'pyrt.camera': 'pyrt/camera',
        'pyrt.geometry': 'pyrt/geometry',
        'pyrt.light': 'pyrt/light',
        'pyrt.material': 'pyrt/material',
        'pyrt.math': 'pyrt/math',
        'pyrt.renderer': 'pyrt/renderer',
        'pyrt.scene': 'pyrt/scene'
    },
    package_data={'pyrt': ['camera/*.py']},
    include_package_data=True,
    version='0.0.1a1',
    description='pyRT - The Python Raytracer',
    long_description=long_description,
    url='https://github.com/martinchristen/pyRT',
    author='Martin Christen',
    author_email='martin.christen@gmail.com',
    license='MIT',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Education',
        'Topic :: Multimedia :: Graphics :: 3D Rendering',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.5',
    ],
    keywords=['raytracing','3d-graphics'],
)


"""
This is the geometric shape description

(renderable object)
"""

from abc import abstractmethod
from ..math import Ray, HitRecord
import uuid


class Shape(object):

    """This is the base class for all geometries"""

    def __init__(self, name: str) -> None:
        self.name = name
        self.material = None
        self.id = str(uuid.uuid4())

    @abstractmethod
    def hit(self, ray: Ray, hitrecord: HitRecord) -> bool:
        """
        :param ray: the ray to check hit
        :param tmin: tmin to test intersection
        :param tmax: tmax to test intersection
        :param hitrecord: the hitrecord which is only valid if there is a hit
        :return: True if there is a hit
        """
        pass

    @abstractmethod
    def hitShadow(self, ray: Ray) -> bool:
        """
        :param ray:
        :param tmin:
        :param tmax:
        :return:
        """
        pass

        #####################
        # TODO: BOUNDING BOX
        #       implement abstract method to retrieve bounding box
        ####################

        ####################
        # TODO: MATERIAL
        #       handle shape material (color, textures, ...)
        ####################



"""
Ray - Handling rays for ray tracing

This is the code for handling 3D rays
"""

from .vec3 import Vec3


class HitRecord(object):
    def __init__(self):
        self.reset()

    def reset(self):
        self.t = None  # intersection parameter of line
        self.point = None  # hit point (in world coordinates)
        self.normal = None  # hit normal (shading normal)
        self.normal_g = None  # hit normal (geometric normal)
        self.color = None  # interpolated color
        self.material = None  # Material
        self.texcoord = None  # texture coordinate at hit point
        self.obj = None  # hit object/geometry or None if not applicable

    def copy(self):
        n = HitRecord()
        n.t = self.t
        if self.point is not None:
            n.point = self.point.copy()
        if self.normal is not None:
            n.normal = self.normal.copy()
        if self.normal_g is not None:
            n.normal_g = self.normal_g.copy()
        if self.color is not None:
            n.color = self.color.copy()

        n.material = self.material # Material is still reference!

        if self.texcoord is not None:
            n.texcoord = self.texcoord.copy()
        n.obj = self.obj

        return n


#-----------------------------------------------------------------------------------------------------------------------


class Ray(object):
    def __init__(self, start: Vec3, direction: Vec3) -> None:
        if type(start) != Vec3 or type(direction) != Vec3:
            print("Error: start and direction must be Vec3")
            return

        self.start = start  # todo: Property
        self.direction = direction  # todo: Property

        # self.invdir = Vec3(1.0 / self.direction.x, 1.0 / self.direction.y, 1.0 / self.direction.z)
        # self.sign = sign(self.invdir)

    def copy(self):
        """
        Create a copy of the reay

        :return: copy of ray
        """
        return Ray(self.start.copy(), self.direction.copy())
